
/*
 *  finddup is a simple utility to display the files and directories
 *  according to their total disk occupancy.
 *
 *  Copyright (c) 2010 Francois Fleuret
 *  Written by Francois Fleuret <francois@fleuret.org>
 *
 *  This file is part of finddup.
 *
 *  finddup is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version 3 as
 *  published by the Free Software Foundation.
 *
 *  finddup is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 *  License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with finddup.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define _BSD_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <locale.h>
#include <getopt.h>
#include <fcntl.h>

#define BUFFER_SIZE 4096

typedef int64_t size_sum_t;

/* Yeah, global variables! */

int ignore_dotfiles = 0; /* 1 means ignore files and directories
                            starting with a dot */

int forced_width = 0; /* -1 means no width limit, strictly positive
                         means limit, 0 means not active */

int forced_height = 0; /* -1 means no height limit, strictly positive
                           means limit, 0 means not active */

int fancy_size_display = 0; /* 1 means to use floating values with K, M and G
                               as units */

int reverse_sorting = 0; /* 1 means to show the large ones first */

int show_top = 0; /* 1 means to show the top of the sorted list
                     instead of the bottom */

/********************************************************************/

/* malloc with error checking.  */

void *safe_malloc(size_t n) {
  void *p = malloc(n);
  if (!p && n != 0) {
    printf("Can not allocate memory: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }
  return p;
}

/********************************************************************/

int ignore_entry(const char *name) {
  return
    strcmp(name, ".") == 0 ||
    strcmp(name, "..") == 0 ||
    (ignore_dotfiles && name[0] == '.');
}

void print_size_sum(size_sum_t s) {
  char tmp[100];
  char *a = tmp + sizeof(tmp)/sizeof(char);
  *(--a) = '\0';
  if(s) {
    while(s) {
      *(--a) = s%10 + '0';
      s /= 10;
    }
  } else {
    *(--a) = '0';
  }
  printf(a);
}

size_sum_t file_or_dir_size(const char *name) {
  DIR *dir;
  struct dirent *dir_e;
  struct stat dummy;
  size_sum_t result;
  char subname[BUFFER_SIZE];

  result = 0;

  if(lstat(name, &dummy) != 0) {
    printf("Can not stat %s: %s\n", name, strerror(errno));
    exit(EXIT_FAILURE);
  }

  if(S_ISLNK(dummy.st_mode)) {
    return 0;
  }

  dir = opendir(name);

  if(dir) {
    while((dir_e = readdir(dir))) {
      if(!ignore_entry(dir_e->d_name)) {
        snprintf(subname, BUFFER_SIZE, "%s/%s", name, dir_e->d_name);
        result += file_or_dir_size(subname);
      }
    }
    closedir(dir);
  } else {
    if(S_ISREG(dummy.st_mode)) {
      result += dummy.st_size;
    }
  }

  return result;
}

/**********************************************************************/

struct file_with_size {
  char *filename;
  size_t size;
  struct file_with_size *next;
  int error;
};

struct file_list {
  struct file_with_size *head, *tail;
};

void create(struct file_list *list, const char *name) {
  list->head = safe_malloc(sizeof(struct file_with_size));
  list->tail = 0;
  list->head->filename = strdup(name);
  list->head->size = file_or_dir_size(name);
  list->head->next = 0;
}

void destroy(struct file_with_size *node) {
  struct file_with_size *next;
  while(node) {
    next = node->next;
    free(node->filename);
    free(node);
    node = next;
  }
}

/**********************************************************************/

int same_size(struct file_with_size *f1, struct file_with_size *f2) {
  return f1->size == f2->size;
}

int same_content(struct file_with_size *f1, struct file_with_size *f2) {
  int fd1, fd2, eq, s1, s2;
  char buffer1[BUFFER_SIZE], buffer2[BUFFER_SIZE];

  fd1 = open(f1->filename, O_RDONLY);
  fd2 = open(f2->filename, O_RDONLY);
  eq = 1;

  if(fd1 >= 0 && fd2 >= 0) {
    while(eq) {
      s1 = read(fd1, buffer1, BUFFER_SIZE);
      if(s1 < 0) {
        f1->error = 1;
        eq = 0;
      }
      s2 = read(fd2, buffer2, BUFFER_SIZE);
      if(s2 < 0) {
        f2->error = 1;
        eq = 0;
      }
      if(s1 > 0 && s2 > 0) {
        if(s1 == s2) {
          eq = (strncmp(buffer1, buffer2, s1) == 0);
        } else {
          fprintf(stderr,
                  "Different read size without error on files of same size.\n");
          exit(EXIT_FAILURE);
        }
      }
    }
  }

  if(fd1 >= 0) {
    close(fd1);
  } else {
    f1->error = 1;
    eq = 0;
  }

  if(fd2 >= 0) {
    close(fd2);
  } else {
    f2->error = 1;
    eq = 0;
  }

  return eq;
}

int same_files(struct file_with_size *f1, struct file_with_size *f2) {
  return same_size(f1, f2) && same_content(f1, f2);
}

/**********************************************************************/

struct file_with_size *scan_directory(struct file_with_size *tail, const char *name) {
  DIR *dir;
  struct dirent *dir_e;
  struct stat dummy;
  struct file_with_size *tmp;
  char subname[BUFFER_SIZE];

  if(lstat(name, &dummy) != 0) {
    fprintf(stderr, "Can not stat %s: %s\n", name, strerror(errno));
    exit(EXIT_FAILURE);
  }

  if(S_ISLNK(dummy.st_mode)) {
    return tail;
  }

  dir = opendir(name);

  if(dir) {
    while((dir_e = readdir(dir))) {
      if(!ignore_entry(dir_e->d_name)) {
        snprintf(subname, BUFFER_SIZE, "%s/%s", name, dir_e->d_name);
        tail = scan_directory(tail, subname);
      }
    }
    closedir(dir);
  } else {
    if(S_ISREG(dummy.st_mode)) {
      tmp = safe_malloc(sizeof(struct file_with_size));
      tmp->next = tail;
      tmp->filename = strdup(name);
      tmp->size = file_or_dir_size(name);
      tail = tmp;
    }
  }

  return tail;
}

void start(const char *dirname1, const char *dirname2) {
  struct file_with_size *list1, *list2;
  struct file_with_size *node;
  list1 = scan_directory(0, dirname1);
  list2 = scan_directory(0, dirname2);

  for(node = list1; node; node = node->next) {
    printf("list1: %s\n", node->filename);
  }

  for(node = list2; node; node = node->next) {
    printf("list2: %s\n", node->filename);
  }

  destroy(list1);
  destroy(list2);
}

/**********************************************************************/

int main(int argc, char **argv) {
  int c;
  struct file_with_size *root;

  root = 0;

  setlocale (LC_ALL, "");

  while (1) {
    c = getopt(argc, argv, "h");
    if (c == -1)
      break;

    switch (c) {

    case 'h':
      printf("Usage: finddup [OPTION]... [FILE]...\n");
      printf("Report bugs and comments to <francois@fleuret.org>\n");
      exit(EXIT_SUCCESS);

      break;

    default:
      exit(EXIT_FAILURE);
    }
  }

  if(optind + 1 < argc) {
    start(argv[optind], argv[optind + 1]);
  } else {
    fprintf(stderr, "%s [OPTIONS] <dir1> <dir2>\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  exit(EXIT_SUCCESS);
}
